@@ -6,6 +6,8 @@
 #include <plog/Initializers/RollingFileInitializer.h>
 #include <plog/Log.h>
 
+#include <dlfcn.h>
+
 #include <fstream>
 #include <iostream>
 
@@ -39,7 +41,6 @@ struct UdfDataBlock {
 
   py::object data(int i, int j) {
     if (i < 0 || i > _dataBlock->numOfRows || j < 0 || j > _dataBlock->numOfCols) {
-      // TODO: choose indexerror exception
       throw py::index_error("out of range for data call");
     }
     SUdfColumn *col = _dataBlock->udfCols[j];
@@ -81,7 +82,6 @@ struct UdfDataBlock {
       case TSDB_DATA_TYPE_BINARY:
         return py::bytes(varDataVal(data), varDataLen(data));
 
-      // TODO: json
       case TSDB_DATA_TYPE_JSON:
       default:
         throw py::type_error("unsupported python udf type");
@@ -157,7 +157,6 @@ struct UdfDataBlock {
             udfColDataSet(col, i, data.get(), false);
             break;
           }
-          // TODO: json
           case TSDB_DATA_TYPE_JSON:
           default: {
             throw py::type_error("unsupported python udf type");
@@ -194,7 +193,7 @@ class PyUdf {
         _funcType(udfInfo->funcType),
         _outputType(udfInfo->outputType),
         _outputLen(udfInfo->outputLen),
-        _bufSize(udfInfo->bufSize) {
+        _bufSize(udfInfo->bufSize) { 
     std::filesystem::path p(_path);
     _module = py::module_::import(p.stem().c_str());
   }
@@ -235,9 +234,9 @@ class PyScalarUdf : public PyUdf {
 
   int32_t scalarProc(SUdfDataBlock *dataBlock, SUdfColumn *resultCol) {
     SUdfColumnData *resultData = &resultCol->colData;
-    UdfDataBlock block(dataBlock);
-    py::object   pyblk = py::cast(&block);
-    py::tuple    pyTuple = _scalarProc(pyblk);
+    UdfDataBlock    block(dataBlock);
+    py::object      pyblk = py::cast(&block);
+    py::tuple       pyTuple = _scalarProc(pyblk);
     if (pyTuple.size() != dataBlock->numOfRows) {
       PLOGE << "scalar udf. row number: " << dataBlock->numOfRows << ", result size: " << pyTuple.size();
       throw std::runtime_error("python udf scalar function shall return each result for each row.");
@@ -338,8 +337,14 @@ class PyAggUdf : public PyUdf {
   }
 
   void aggFinish(SUdfInterBuf *buf, SUdfInterBuf *resultData) {
-    py::bytes  pybuf(buf->buf, buf->bufLen);
-    py::object obj = _aggFinish(pybuf);
+    py::object pyobjBuf;
+    if (buf->numOfResult == 0) {
+      pyobjBuf = py::none();
+    } else {
+      py::bytes pybuf(buf->buf, buf->bufLen);
+      pyobjBuf = pybuf;
+    }
+    py::object obj = _aggFinish(pyobjBuf);
     if (obj.is_none()) {
       resultData->numOfResult = 0;
     } else {
@@ -399,7 +404,6 @@ class PyAggUdf : public PyUdf {
           resultData->numOfResult = 1;
           break;
         }
-        // TODO: json
         case TSDB_DATA_TYPE_JSON:
         default: {
           throw py::type_error("unsupported python udf type");
@@ -427,6 +431,7 @@ PyUdf *PyUdf::createPyUdf(const SScriptUdfInfo *udfInfo) {
 }
 
 int32_t doPyUdfInit(SScriptUdfInfo *udf, void **pUdfCtx) {
+  PLOGD << "python udf init. path: " << udf->path << ", name: " << udf->name; 
   try {
     PyUdf *pyUdf = PyUdf::createPyUdf(udf);
     pyUdf->loadFunctions();
@@ -435,10 +440,12 @@ int32_t doPyUdfInit(SScriptUdfInfo *udf, void **pUdfCtx) {
     PLOGE << "call pyUdf init function. error " << e.what();
     return TSDB_UDF_PYTHON_EXEC_FAILURE;
   }
+  PLOGI << "python udf init. name " << udf->name << ". context: " << static_cast<void*>(*pUdfCtx);
   return 0;
 }
 
 int32_t doPyUdfDestroy(void *udfCtx) {
+  PLOGD << "python udf destory. context: " << static_cast<void*>(udfCtx); 
   try {
     PyUdf *pyUdf = static_cast<PyUdf *>(udfCtx);
     delete pyUdf;
@@ -450,37 +457,40 @@ int32_t doPyUdfDestroy(void *udfCtx) {
 }
 
 int32_t doPyUdfScalarProc(SUdfDataBlock *block, SUdfColumn *resultCol, void *udfCtx) {
+  PLOGD << "call pyUdfScalar proc function. context " << static_cast<void*>(udfCtx) << ". rows: " << block->numOfRows;
   try {
     PyUdf       *pyUdf = static_cast<PyUdf *>(udfCtx);
     PyScalarUdf *pyScalarUdf = dynamic_cast<PyScalarUdf *>(pyUdf);
     pyScalarUdf->scalarProc(block, resultCol);
     return 0;
   } catch (std::exception &e) {
-    PLOGE << "call pyUdfScalar proc function. error " << e.what();
+    PLOGE << "call pyUdfScalar proc function. context " << static_cast<void*>(udfCtx) << ". error: " << e.what();
     return TSDB_UDF_PYTHON_EXEC_FAILURE;
   }
   return 0;
 }
 
 int32_t doPyUdfAggStart(SUdfInterBuf *buf, void *udfCtx) {
+  PLOGD << "call pyUdfAgg start function. context " << static_cast<void*>(udfCtx);
   try {
     PyUdf    *pyUdf = static_cast<PyUdf *>(udfCtx);
     PyAggUdf *pyAggUdf = dynamic_cast<PyAggUdf *>(pyUdf);
     pyAggUdf->aggStart(buf);
   } catch (std::exception &e) {
-    PLOGE << "call pyUdfAgg start function. error " << e.what();
+    PLOGE << "call pyUdfAgg start function. context " << static_cast<void*>(udfCtx) << " .error: " << e.what();
     return TSDB_UDF_PYTHON_EXEC_FAILURE;
   }
   return 0;
 }
 
 int32_t doPyUdfAggProc(SUdfDataBlock *block, SUdfInterBuf *interBuf, SUdfInterBuf *newInterBuf, void *udfCtx) {
+  PLOGD << "call pyAggUdf proc function. context " << static_cast<void*>(udfCtx) << ". rows: " << block->numOfRows;
   try {
     PyUdf    *pyUdf = static_cast<PyUdf *>(udfCtx);
     PyAggUdf *pyAggUdf = dynamic_cast<PyAggUdf *>(pyUdf);
     pyAggUdf->aggProc(block, interBuf, newInterBuf);
   } catch (std::exception &e) {
-    PLOGE << "call pyAggUdf proc function. error " << e.what();
+    PLOGE << "call pyAggUdf proc function. context " << static_cast<void*>(udfCtx) << ". error " << e.what();
     return TSDB_UDF_PYTHON_EXEC_FAILURE;
   }
   return 0;
@@ -501,12 +511,13 @@ int32_t doPyUdfAggMerge(SUdfInterBuf *inputBuf1, SUdfInterBuf *inputBuf2, SUdfIn
 }
 
 int32_t doPyUdfAggFinish(SUdfInterBuf *buf, SUdfInterBuf *resultData, void *udfCtx) {
+  PLOGD << "call pyAggUdf finish function. context " << static_cast<void*>(udfCtx);
   try {
     PyUdf    *pyUdf = static_cast<PyUdf *>(udfCtx);
     PyAggUdf *pyAggUdf = dynamic_cast<PyAggUdf *>(pyUdf);
     pyAggUdf->aggFinish(buf, resultData);
   } catch (std::exception &e) {
-    PLOGE << "call pyAggUdf finish function. eror " << e.what();
+    PLOGE << "call pyAggUdf finish function. context " << static_cast<void*>(udfCtx) << ". error " << e.what();
     return TSDB_UDF_PYTHON_EXEC_FAILURE;
   }
   return 0;
@@ -519,6 +530,7 @@ std::vector<std::string> resplit(const std::string &s, const std::regex &sep_reg
 }
 
 int32_t doPyOpen(SScriptUdfEnvItem *items, int numItems) {
+  PLOGI << "python udf plugin open";
   if (Py_IsInitialized() == 1) {
     return TSDB_UDF_PYTHON_WRONG_STATE;
   }
@@ -535,26 +547,27 @@ int32_t doPyOpen(SScriptUdfEnvItem *items, int numItems) {
     }
     auto taosPyUdf = py::module::import("taospyudf");
   } catch (std::exception &e) {
-    PLOGE << "pyOpen error. " << e.what();
+    PLOGE << "python udf plugin open error. " << e.what();
     return TSDB_UDF_PYTHON_EXEC_FAILURE;
   }
   return 0;
 }
 
 int32_t doPyClose() {
+  PLOGI << "python udf plugin close.";
   if (Py_IsInitialized() == 0) {
     return TSDB_UDF_PYTHON_WRONG_STATE;
   }
   try {
     py::finalize_interpreter();
   } catch (std::exception &e) {
-    std::cerr << e.what() << std::endl;
+    PLOGE << "python udf plugin close. " << e.what();
 
     return TSDB_UDF_PYTHON_EXEC_FAILURE;
   }
   return 0;
 }
-
+// thread pool from https://maidamai0.github.io/post/a-simple-thread-pool/
 class ThreadPool {
   using task_type = std::function<void()>;
 
@@ -578,7 +591,7 @@ class ThreadPool {
           task();
         }
       }));
-      PLOGI << "worker #" << workers_.back().get_id() << " started";
+      PLOGI << "python udf worker #" << workers_.back().get_id() << " started";
     }
   }
 
@@ -664,6 +677,8 @@ int32_t pyUdfAggFinish(SUdfInterBuf *buf, SUdfInterBuf *resultData, void *udfCtx
 }
 
 int32_t pyOpen(SScriptUdfEnvItem *items, int numItems) {
+  dlopen("libtaospyudf.so", RTLD_LAZY | RTLD_GLOBAL);
+
   std::string logPath("/tmp/");
   for (int i = 0; i < numItems; ++i) {
     if (std::string_view(items[i].name) == std::string_view("LOGDIR")) {
@@ -672,8 +687,8 @@ int32_t pyOpen(SScriptUdfEnvItem *items, int numItems) {
     }
   }
   logPath += std::string("/taospyudf.log");
-  plog::init(plog::debug, logPath.c_str(), 50 * 1024 * 1024, 5);
-  PLOGI << "taos python udf open";
+  plog::init(plog::info, logPath.c_str(), 50 * 1024 * 1024, 5);
+  PLOGI << "taos python udf plugin open";
   // only one caller
   pythonCaller = new ThreadPool(1);
   auto f = pythonCaller->enqueue(doPyOpen, items, numItems);
@@ -685,13 +700,7 @@ int32_t pyClose() {
   int32_t ret = f.get();
   delete pythonCaller;
   pythonCaller = nullptr;
-  PLOGI << "taos python udf close";
+  dlopen("libtaospyudf.so", RTLD_LAZY | RTLD_GLOBAL);
+  PLOGI << "taos python udf plugin close";
   return ret;
 }
-// TODO: catch exception in pyInterfaceFunction
-// TODO: pyInterfacefucntion shall be externc
-// TODO: scalar proc udf column result meta set in udfd, remove it from sample libbit_and
-// TODO: so and install with taos-py-connector
-// TODO: optional function
-// TODO: log library
-// TODO: error python exception -2, unknown, change to app error
